# Copyright (C) 2011-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import plistlib


def parse_bookmark_data(data):
    """Return iterator for bookmarks of the form (url, name).
    Bookmarks are not sorted.
    """
    return parse_plist(get_plist_data_from_string(data))


def get_plist_data_from_string(data):
    """Parse plist data for a string."""
    try:
        return plistlib.loads(data)
    except Exception:
        # not parseable (eg. not well-formed)
        return {}


# some key strings
KEY_URLSTRING = 'URLString'
KEY_URIDICTIONARY = 'URIDictionary'
KEY_CHILDREN = 'Children'
KEY_WEBBOOKMARKTYPE = 'WebBookmarkType'


def parse_plist(entry):
    """Parse a XML dictionary entry."""
    if is_leaf(entry):
        url = entry[KEY_URLSTRING]
        title = entry[KEY_URIDICTIONARY].get('title', url)
        yield (url, title)
    elif has_children(entry):
        for child in entry[KEY_CHILDREN]:
            yield from parse_plist(child)


def is_leaf(entry):
    """Return true if plist entry is an URL entry."""
    return entry.get(KEY_WEBBOOKMARKTYPE) == 'WebBookmarkTypeLeaf'


def has_children(entry):
    """Return true if plist entry has children."""
    return entry.get(KEY_WEBBOOKMARKTYPE) == 'WebBookmarkTypeList'



# Copyright (C) 2022 Stefan Fisk
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
srcset attribute parser
"""


_TAB = '\u0009'
_LF = '\u000A'
_FF = '\u000C'
_CR = '\u000D'
_SPACE = '\u0020'
_COMMA = '\u002C'
_LEFT_PARENTHESIS = '\u0028'
_RIGHT_PARENTHESIS = '\u0029'


_WHITESPACE = {_TAB,  _LF,  _FF,  _CR,  _SPACE}
_WHITESPACE_OR_COMMA = _WHITESPACE | {_COMMA}


def parse_srcset(input):
    """
    Parse HTML srcset

    Based on WhatWG HTML standard § 4.8.4.3.10 Parsing a srcset attribute,
    but does not parse or validate descriptors.

    https://html.spec.whatwg.org/multipage/images.html#parse-a-srcset-attribute
    """

    input_end = len(input)
    position = 0
    urls = []

    while position < input_end:
        # 4. Splitting loop: Collect a sequence of code points that are ASCII
        # whitespace or U+002C COMMA characters from input given position.
        while position < input_end and input[position] in _WHITESPACE_OR_COMMA:
            position += 1

        # 5. If position is past the end of input, return candidates.
        if position >= input_end:
            return urls

        # 6. Collect a sequence of code points that are not ASCII
        # whitespace from input given position, and let that be url.
        url_start = position
        while position < input_end and input[position] not in _WHITESPACE:
            position += 1
        url_end = position

        # 8, If url ends with U+002C (,), then:
        if input[url_end - 1] == _COMMA:
            # Remove all trailing U+002C COMMA characters from url.
            while url_end > url_start and input[url_end - 1] == _COMMA:
                url_end -= 1
        else:
            # This is a shortened version of 1–4 that simply skips the
            # descriptors
            while position < input_end:
                if input[position] == _LEFT_PARENTHESIS:
                    # Skip until first closing parenthesis
                    while (position < input_end and input[position] !=
                            _RIGHT_PARENTHESIS):
                        position += 1
                elif input[position] == _COMMA:
                    break

                position += 1

        # 9-15 is parsing and validation of the descriptors, which we ignore

        # If we found an URL
        if url_end > url_start:
            urls.append(input[url_start:url_end])

    return urls



# Copyright (C) 2006-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test config parsing.
"""

import os
from re import Pattern
import linkcheck.configuration

from .. import TestBase


def get_file(filename=None):
    """Get file name located within 'data' directory."""
    directory = os.path.join("tests", "configuration", "data")
    if filename:
        return os.path.join(directory, filename)
    return directory


class TestConfig(TestBase):
    """Test configuration parsing."""

    def test_confparse(self):
        # Tests must not use the default value only
        config = linkcheck.configuration.Configuration()
        files = [get_file("config0.ini")]
        config.read(files)
        config.sanitize()
        # checking section
        for scheme in ("http", "https", "ftp"):
            self.assertTrue(scheme in config["allowedschemes"])
        self.assertEqual(config["threads"], 5)
        self.assertEqual(config["timeout"], 42)
        self.assertEqual(config["aborttimeout"], 99)
        self.assertEqual(config["recursionlevel"], 1)
        self.assertEqual(config["cookiefile"], "blablabla")
        self.assertEqual(config["useragent"], "Example/0.0")
        self.assertEqual(config["debugmemory"], 1)
        self.assertEqual(config["localwebroot"], "foo")
        self.assertEqual(config["sslverify"], "/path/to/cacerts.crt")
        self.assertEqual(config["maxnumurls"], 1000)
        self.assertEqual(config["maxrequestspersecond"], 0.1)
        self.assertEqual(config["maxrunseconds"], 1)
        self.assertEqual(config["maxfilesizeparse"], 100)
        self.assertEqual(config["maxfilesizedownload"], 100)
        self.assertEqual(config["resultcachesize"], 9999)
        # filtering section
        patterns = [x["pattern"].pattern for x in config["externlinks"]]
        for prefix in ("ignore_", "nofollow_"):
            for suffix in ("1", "2"):
                key = f"{prefix}imadoofus{suffix}"
                self.assertTrue(key in patterns)
        for key in ("url-unicode-domain",):
            self.assertTrue(key in config["ignorewarnings"])
        for parts in config["ignorewarningsforurls"]:
            self.assertEqual(len(parts), 2)
            for part in parts:
                self.assertTrue(isinstance(part, Pattern))
        self.assertTrue(config["ignorewarningsforurls"][0][1].search(
            "http-redirected"
        ))
        self.assertTrue(config["ignorewarningsforurls"][1][0].search(
            "https://example.com/another-url"
        ))
        self.assertEqual(len(config["ignoreerrors"]), 2)
        for parts in config["ignoreerrors"]:
            self.assertEqual(len(parts), 2)
            for part in parts:
                self.assertTrue(isinstance(part, Pattern))
        self.assertTrue(config["ignoreerrors"][0][1].search(
            "404 Not Found"
        ))
        self.assertTrue(config["ignoreerrors"][1][0].search(
            "mailto:foo"
        ))
        self.assertTrue(config["checkextern"])
        # authentication section
        patterns = [x["pattern"].pattern for x in config["authentication"]]
        for suffix in ("1", "2"):
            key = "imadoofus%s" % suffix
            self.assertTrue(key in patterns)
        self.assertTrue("http://www.example.com/" in patterns)
        self.assertTrue("http://www.example.com/nopass" in patterns)
        self.assertEqual(config["loginurl"], "http://www.example.com/")
        self.assertEqual(config["loginuserfield"], "mylogin")
        self.assertEqual(config["loginpasswordfield"], "mypassword")
        self.assertEqual(config["loginextrafields"]["name1"], "value1")
        self.assertEqual(config["loginextrafields"]["name 2"], "value 2")
        self.assertEqual(len(config["loginextrafields"]), 2)
        # output section
        self.assertTrue(linkcheck.log.is_debug(linkcheck.LOG_THREAD))
        self.assertFalse(config["status"])
        self.assertTrue(
            isinstance(config["logger"], linkcheck.logger.customxml.CustomXMLLogger)
        )
        self.assertTrue(config["verbose"])
        self.assertTrue(config["warnings"])
        self.assertFalse(config["quiet"])
        self.assertEqual(len(config["fileoutput"]), 8)
        # plugins
        for plugin in (
            "AnchorCheck",
            "CssSyntaxCheck",
            "HtmlSyntaxCheck",
            "LocationInfo",
            "RegexCheck",
            "SslCertificateCheck",
            "VirusCheck",
            "HttpHeaderInfo",
        ):
            self.assertTrue(plugin in config["enabledplugins"])
        # text logger section
        self.assertEqual(config["text"]["filename"], "imadoofus.txt")
        self.assertEqual(config["text"]["parts"], ["realurl"])
        self.assertEqual(config["text"]["encoding"], "utf-8")
        self.assertEqual(config["text"]["wraplength"], "80")
        self.assertEqual(config["text"]["colorparent"], "blink;red")
        self.assertEqual(config["text"]["colorurl"], "blink;red")
        self.assertEqual(config["text"]["colorname"], "blink;red")
        self.assertEqual(config["text"]["colorreal"], "blink;red")
        self.assertEqual(config["text"]["colorbase"], "blink;red")
        self.assertEqual(config["text"]["colorvalid"], "blink;red")
        self.assertEqual(config["text"]["colorinvalid"], "blink;red")
        self.assertEqual(config["text"]["colorinfo"], "blink;red")
        self.assertEqual(config["text"]["colorwarning"], "blink;red")
        self.assertEqual(config["text"]["colordltime"], "blink;red")
        self.assertEqual(config["text"]["colorreset"], "blink;red")
        # gml logger section
        self.assertEqual(config["gml"]["filename"], "imadoofus.gml")
        self.assertEqual(config["gml"]["parts"], ["realurl"])
        self.assertEqual(config["gml"]["encoding"], "utf-8")
        # dot logger section
        self.assertEqual(config["dot"]["filename"], "imadoofus.dot")
        self.assertEqual(config["dot"]["parts"], ["realurl"])
        self.assertEqual(config["dot"]["encoding"], "utf-8")
        # csv logger section
        self.assertEqual(config["csv"]["filename"], "imadoofus.csv")
        self.assertEqual(config["csv"]["parts"], ["realurl"])
        self.assertEqual(config["csv"]["encoding"], "utf-8")
        self.assertEqual(config["csv"]["separator"], ";")
        self.assertEqual(config["csv"]["quotechar"], "'")
        # sql logger section
        self.assertEqual(config["sql"]["filename"], "imadoofus.sql")
        self.assertEqual(config["sql"]["parts"], ["realurl"])
        self.assertEqual(config["sql"]["encoding"], "utf-8")
        self.assertEqual(config["sql"]["separator"], ";")
        self.assertEqual(config["sql"]["dbname"], "linksdb")
        # html logger section
        self.assertEqual(config["html"]["filename"], "imadoofus.html")
        self.assertEqual(config["html"]["parts"], ["realurl"])
        self.assertEqual(config["html"]["encoding"], "utf-8")
        self.assertEqual(config["html"]["colorbackground"], "#ff0000")
        self.assertEqual(config["html"]["colorurl"], "#ff0000")
        self.assertEqual(config["html"]["colorborder"], "#ff0000")
        self.assertEqual(config["html"]["colorlink"], "#ff0000")
        self.assertEqual(config["html"]["colorwarning"], "#ff0000")
        self.assertEqual(config["html"]["colorerror"], "#ff0000")
        self.assertEqual(config["html"]["colorok"], "#ff0000")
        # failures logger section
        self.assertEqual(config["failures"]["filename"], "failures")
        self.assertEqual(config["failures"]["encoding"], "utf-8")
        # xml logger section
        self.assertEqual(config["xml"]["filename"], "imadoofus.xml")
        self.assertEqual(config["xml"]["parts"], ["realurl"])
        self.assertEqual(config["xml"]["encoding"], "utf-8")
        # gxml logger section
        self.assertEqual(config["gxml"]["filename"], "imadoofus.gxml")
        self.assertEqual(config["gxml"]["parts"], ["realurl"])
        self.assertEqual(config["gxml"]["encoding"], "utf-8")

    def test_confparse_error1(self):
        config = linkcheck.configuration.Configuration()
        files = [get_file("config1.ini")]
        self.assertRaises(linkcheck.LinkCheckerError, config.read, files)

    def test_confparse_error2(self):
        config = linkcheck.configuration.Configuration()
        files = [get_file("config2.ini")]
        self.assertRaises(linkcheck.LinkCheckerError, config.read, files)

    def test_confparse_deprecated(self):
        config = linkcheck.configuration.Configuration()
        files = [get_file("config3.ini")]
        config.read(files)
        config.sanitize()
        # blacklist logger section
        self.assertEqual(config["failures"]["filename"], "blacklist")
        self.assertEqual(config["failures"]["encoding"], "utf-8")

    def test_confparse_empty(self):
        config = linkcheck.configuration.Configuration()
        files = [get_file("config.empty")]
        self.assertRaises(linkcheck.LinkCheckerError, config.read, files)

    def test_confparse_missing(self):
        config = linkcheck.configuration.Configuration()
        files = [get_file("no_such_config")]
        self.assertRaises(linkcheck.LinkCheckerError, config.read, files)



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test ignoring of errors.
"""

from re import compile as re_compile

from tests import need_network
from . import LinkCheckTest


class TestIgnoreErrors(LinkCheckTest):
    """
    Test whether ignoring of errors per URL works.
    """

    def _test(self, url, url_regex, msg_regex, valid):
        """ Shorthand for various tests of ignoring errors. """
        confargs = {
            "ignoreerrors": [
                (re_compile(url_regex), re_compile(msg_regex))
            ]
        }
        resultlines = [
            "url %s" % url,
            "cache key %s" % url,
            "real url %s" % url,
            "valid" if valid else "error",
        ]
        self.direct(url, resultlines, confargs=confargs)

    @need_network
    def test_no_error(self):
        """ Test that unmatched errors are not ignored. """
        self._test("mailto:good@example.com", "", "", True)
        self._test("mailto:good@example.com", "^$", "", True)
        self._test("mailto:good@example.com", "^$", "^no-match$", True)
        self._test("mailto:good@example.com",
                   r"^mailto:good@example\.com$", "", True)
        self._test("mailto:good@example.com",
                   r"^mailto:good@example\.com$", "^$", True)

    def test_url_regex(self):
        """ Test that URLs are properly matched. """
        self._test("mailto:foo", r"^$", "", False)
        self._test("mailto:foo", r"", "", True)
        self._test("mailto:foo", r"^mailto:foo$", "", True)
        self._test("mailto:foobar", r"^mailto:foo", "", True)

    def test_msg_regex(self):
        """ Test that error messages are properly matched. """
        self._test("mailto:foo", r"^mailto:foo$", "^$", False)
        self._test("mailto:foo", r"^mailto:foo$", "", True)
        self._test("mailto:foo", r"^mailto:foo$",
                   r"^Missing `@' in mail address `foo'.$", True)

    @need_network
    def test_internet(self):
        """ Test a few well-known Internet URLs. """
        self._test("https://linkchecker.github.io/does-not-exist",
                   r"^https://linkchecker.github.io/.+$", "^404", True)
        self._test("http://does-not-exist.example.com",
                   r"example.com", "^ConnectionError", True)



# Copyright (C) 2006-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Helpers for console output.
"""
import sys
import os
import time
from .. import i18n, configuration, strformat, better_exchook2

# Output to stdout and stderr, encoded with the default encoding
stderr = i18n.get_encoded_writer(out=sys.stderr)
stdout = i18n.get_encoded_writer()


class StatusLogger:
    """Standard status logger. Default output is stderr."""

    def __init__(self, fd=stderr):
        """Save file descriptor for logging."""
        self.fd = fd

    def log_status(self, checked, in_progress, queue, duration, num_urls):
        """Write status message to file descriptor."""
        msg = _n("%2d thread active", "%2d threads active", in_progress) % in_progress
        self.write("%s, " % msg)
        msg = _n("%5d link queued", "%5d links queued", queue) % queue
        self.write("%s, " % msg)
        msg = _n("%4d link", "%4d links", checked) % checked
        self.write("%s" % msg)
        msg = _n("%3d URL", "%3d URLs", num_urls) % num_urls
        self.write(" in %s checked, " % msg)
        msg = _("runtime %s") % strformat.strduration_long(duration)
        self.writeln(msg)
        self.flush()

    def write(self, msg):
        """Write message to file descriptor."""
        self.fd.write(msg)

    def writeln(self, msg):
        """Write status message and line break to file descriptor."""
        self.fd.write(f"{msg}{os.linesep}")

    def flush(self):
        """Flush file descriptor."""
        self.fd.flush()


def internal_error(out=stderr, etype=None, evalue=None, tb=None):
    """Print internal error message (output defaults to stderr)."""
    print(os.linesep, file=out)
    print(_("""********** Oops, I did it again. *************

You have found an internal error in LinkChecker. Please write a bug report
at %s
and include the following information:
- the URL or file you are testing
- the system information below

When using the commandline client:
- your commandline arguments and any custom configuration files.
- the output of a debug run with option "-Dall"

Not disclosing some of the information above due to privacy reasons is ok.
I will try to help you nonetheless, but you have to give me something
I can work with ;) .
""") % configuration.SupportUrl, file=out)
    if etype is None:
        etype = sys.exc_info()[0]
    if evalue is None:
        evalue = sys.exc_info()[1]
    if tb is None:
        tb = sys.exc_info()[2]
    better_exchook2.better_exchook(etype, evalue, tb, out=out)
    print_app_info(out=out)
    print_proxy_info(out=out)
    print_locale_info(out=out)
    print(
        os.linesep,
        _("******** LinkChecker internal error, over and out ********"),
        file=out,
    )


def print_env_info(key, out=stderr):
    """If given environment key is defined, print it out."""
    value = os.getenv(key)
    if value is not None:
        print(key, "=", repr(value), file=out)


def print_proxy_info(out=stderr):
    """Print proxy info."""
    for key in ("http_proxy", "https_proxy", "curl_ca_bundle", "no_proxy"):
        print_env_info(key, out=out)


def print_locale_info(out=stderr):
    """Print locale info."""
    for key in ("LANGUAGE", "LC_ALL", "LC_CTYPE", "LANG"):
        print_env_info(key, out=out)


# Environment variables influencing the interpreter execution
# See python(1) man page.
PYTHON_ENV_VARS = (
    'PYTHONHOME',
    'PYTHONPATH',
    'PYTHONSTARTUP',
    'PYTHONY2K',
    'PYTHONOPTIMIZE',
    'PYTHONDEBUG',
    'PYTHONDONTWRITEBYTECODE',
    'PYTHONINSPECT',
    'PYTHONIOENCODING',
    'PYTHONNOUSERSITE',
    'PYTHONUNBUFFERED',
    'PYTHONVERBOSE',
    'PYTHONWARNINGS',
    'PYTHONHASHSEED',
)


def print_app_info(out=stderr):
    """Print system and application info (output defaults to stderr)."""
    print(_("System info:"), file=out)
    print(configuration.App, file=out)
    print(_("Released on:"), configuration.ReleaseDate, file=out)
    print(
        _("Python %(version)s on %(platform)s")
        % {"version": sys.version, "platform": sys.platform},
        file=out,
    )
    for key in PYTHON_ENV_VARS:
        print_env_info(key, out=out)
    print(configuration.get_modules_info(), file=out)
    stime = strformat.strtime(time.time())
    print(_("Local time:"), stime, file=out)
    print(_("sys.argv:"), sys.argv, file=out)


def print_version(out=stdout):
    """Print the program version (output defaults to stdout)."""
    print(configuration.App, _("released"), configuration.ReleaseDate, file=out)
    print(configuration.Copyright, file=out)



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test http checking.
"""
from .httpserver import HttpServerTest
from tests import need_network


class TestHttpMisc(HttpServerTest):
    """Test http:// misc link checking."""

    @need_network
    def test_html_internet(self):
        self.swf_test()
        self.file_test("sitemap.xml")

    def test_html(self):
        self.file_test("sitemapindex.xml")

    def swf_test(self):
        url = self.get_url("test.swf")
        resultlines = [
            "url %s" % url,
            "cache key %s" % url,
            "real url %s" % url,
            "valid",
            "url http://www.example.org/",
            "cache key http://www.example.org/",
            "real url http://www.example.org/",
            "valid",
        ]
        self.direct(url, resultlines, recursionlevel=1)



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Handle a queue of URLs to check.
"""
import threading
import collections
from time import time as _time
from .. import log, LOG_CACHE


class Timeout(Exception):
    """Raised by join()"""

    pass


class Empty(Exception):
    """Exception raised by get()."""

    pass


NUM_PUTS_CLEANUP = 10000


class UrlQueue:
    """A queue supporting several consumer tasks. The task_done() idea is
    from the Python 2.5 implementation of Queue.Queue()."""

    def __init__(self, max_allowed_urls=None):
        """Initialize the queue state and task counters."""
        # Note: don't put a maximum size on the queue since it would
        # lead to deadlocks when all worker threads called put().
        self.queue = collections.deque()
        # mutex must be held whenever the queue is mutating.  All methods
        # that acquire mutex must release it before returning.  mutex
        # is shared between the two conditions, so acquiring and
        # releasing the conditions also acquires and releases mutex.
        self.mutex = threading.Lock()
        # Notify not_empty whenever an item is added to the queue; a
        # thread waiting to get is notified then.
        self.not_empty = threading.Condition(self.mutex)
        self.all_tasks_done = threading.Condition(self.mutex)
        self.unfinished_tasks = 0
        self.finished_tasks = 0
        self.in_progress = 0
        self.shutdown = False
        # Each put() decreases the number of allowed puts.
        # This way we can restrict the number of URLs that are checked.
        if max_allowed_urls is not None and max_allowed_urls <= 0:
            raise ValueError(
                "Non-positive number of allowed URLs: %d" % max_allowed_urls
            )
        self.max_allowed_urls = max_allowed_urls
        self.num_puts = 0

    def qsize(self):
        """Return the approximate size of the queue (not reliable!)."""
        with self.mutex:
            return len(self.queue)

    def empty(self):
        """Return True if the queue is empty, False otherwise.
        Result is thread-safe, but not reliable since the queue could have
        been changed before the result is returned!"""
        with self.mutex:
            return self._empty()

    def _empty(self):
        """Return True if the queue is empty, False otherwise.
        Not thread-safe!"""
        return not self.queue

    def get(self, timeout=None):
        """Get first not-in-progress url from the queue and
        return it. If no such url is available return None.
        """
        with self.not_empty:
            return self._get(timeout)

    def _get(self, timeout):
        """Non thread-safe utility function of self.get() doing the real
        work."""
        if timeout is None:
            while self._empty():
                self.not_empty.wait()
        else:
            if timeout < 0:
                raise ValueError("'timeout' must be a positive number")
            endtime = _time() + timeout
            while self._empty():
                remaining = endtime - _time()
                if remaining <= 0.0:
                    raise Empty()
                self.not_empty.wait(remaining)
        self.in_progress += 1
        return self.queue.popleft()

    def put(self, item):
        """Put an item into the queue.
        Block if necessary until a free slot is available.
        """
        with self.mutex:
            self._put(item)
            self.not_empty.notify()

    def _put(self, url_data):
        """Put URL in queue, increase number of unfinished tasks."""
        if self.shutdown or self.max_allowed_urls == 0:
            return
        key = url_data.cache_url
        cache = url_data.aggregate.result_cache
        if cache.has_result(key):
            log.debug(LOG_CACHE, "skipping %s, %s already cached", url_data.url, key)
            return
        log.debug(LOG_CACHE, "queueing %s", url_data.url)
        if url_data.has_result:
            self.queue.appendleft(url_data)
        else:
            assert key is not None, "no result for None key: %s" % url_data
            if self.max_allowed_urls is not None:
                self.max_allowed_urls -= 1
            self.num_puts += 1
            if self.num_puts >= NUM_PUTS_CLEANUP:
                self.cleanup()
            self.queue.append(url_data)
        self.unfinished_tasks += 1
        # add none value to cache to prevent checking this url multiple times
        cache.add_result(key, None)

    def cleanup(self):
        """Move cached elements to top."""
        self.num_puts = 0
        cached = []
        for i, url_data in enumerate(self.queue):
            key = url_data.cache_url
            cache = url_data.aggregate.result_cache
            if cache.has_non_empty_result(key):
                cached.append(i)
        for pos in cached:
            self._move_to_top(pos)

    def _move_to_top(self, pos):
        """Move element at given position to top of queue."""
        if pos > 0:
            self.queue.rotate(-pos)
            item = self.queue.popleft()
            self.queue.rotate(pos)
            self.queue.appendleft(item)

    def task_done(self, url_data):
        """
        Indicate that a formerly enqueued task is complete.

        Used by Queue consumer threads.  For each get() used to fetch a task,
        a subsequent call to task_done() tells the queue that the processing
        on the task is complete.

        If a join() is currently blocking, it will resume when all items
        have been processed (meaning that a task_done() call was received
        for every item that had been put() into the queue).

        Raises a ValueError if called more times than there were items
        placed in the queue.
        """
        with self.all_tasks_done:
            log.debug(LOG_CACHE, "task_done %s", url_data.url)
            self.finished_tasks += 1
            self.unfinished_tasks -= 1
            self.in_progress -= 1
            if self.unfinished_tasks <= 0:
                if self.unfinished_tasks < 0:
                    raise ValueError('task_done() called too many times')
                self.all_tasks_done.notify_all()

    def join(self, timeout=None):
        """Blocks until all items in the Queue have been gotten and processed.

        The count of unfinished tasks goes up whenever an item is added to the
        queue. The count goes down whenever a consumer thread calls task_done()
        to indicate the item was retrieved and all work on it is complete.

        When the count of unfinished tasks drops to zero, join() unblocks.
        """
        with self.all_tasks_done:
            if timeout is None:
                while self.unfinished_tasks:
                    self.all_tasks_done.wait()
            else:
                if timeout < 0:
                    raise ValueError("'timeout' must be a positive number")
                endtime = _time() + timeout
                while self.unfinished_tasks:
                    remaining = endtime - _time()
                    if remaining <= 0.0:
                        raise Timeout()
                    self.all_tasks_done.wait(remaining)

    def do_shutdown(self):
        """Shutdown the queue by not accepting any more URLs."""
        with self.mutex:
            unfinished = self.unfinished_tasks - len(self.queue)
            self.queue.clear()
            if unfinished <= 0:
                if unfinished < 0:
                    raise ValueError('shutdown is in error')
                self.all_tasks_done.notify_all()
            self.unfinished_tasks = unfinished
            self.shutdown = True

    def status(self):
        """Get tuple (finished tasks, in progress, queue size)."""
        # no need to acquire self.mutex since the numbers are unreliable anyways.
        return (self.finished_tasks, self.in_progress, len(self.queue))



# Copyright (C) 2009-2010 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test ftpparse routine.
"""

import unittest
from linkcheck.ftpparse import ftpparse

patterns = (
    # EPLF format
    # http://pobox.com/~djb/proto/eplf.html
    (
        "+i8388621.29609,m824255902,/,\tdev",
        dict(name="dev", tryretr=False, trycwd=True),
    ),
    (
        "+i8388621.44468,m839956783,r,s10376,\tRFCEPLF",
        dict(name="RFCEPLF", tryretr=True, trycwd=False),
    ),
    # UNIX-style listing, without inum and without blocks
    (
        "-rw-r--r--   1 root     other        531 Jan 29 03:26 README",
        dict(name="README", tryretr=True, trycwd=False),
    ),
    (
        "dr-xr-xr-x   2 root     other        512 Apr  8  1994 etc",
        dict(name="etc", tryretr=False, trycwd=True),
    ),
    (
        "dr-xr-xr-x   2 root     512 Apr  8  1994 etc",
        dict(name="etc", tryretr=False, trycwd=True),
    ),
    (
        "lrwxrwxrwx   1 root     other          7 Jan 25 00:17 bin -> usr/bin",
        dict(name="usr/bin", tryretr=True, trycwd=True),
    ),
    # Also produced by Microsoft's FTP servers for Windows:
    (
        "----------   1 owner    group         1803128 Jul 10 10:18 ls-lR.Z",
        dict(name="ls-lR.Z", tryretr=True, trycwd=False),
    ),
    (
        "d---------   1 owner    group               0 May  9 19:45 Softlib",
        dict(name="Softlib", tryretr=False, trycwd=True),
    ),
    # Also WFTPD for MS-DOS:
    (
        "-rwxrwxrwx   1 noone    nogroup      322 Aug 19  1996 message.ftp",
        dict(name="message.ftp", tryretr=True, trycwd=False),
    ),
    # Also NetWare:
    (
        "d [R----F--] supervisor            512       Jan 16 18:53    login",
        dict(name="login", tryretr=False, trycwd=True),
    ),
    (
        "- [R----F--] rhesus             214059       Oct 20 15:27    cx.exe",
        dict(name="cx.exe", tryretr=True, trycwd=False),
    ),
    # Also NetPresenz for the Mac:
    (
        "-------r--         326  1391972  1392298 Nov 22  1995 MegaPhone.sit",
        dict(name="MegaPhone.sit", tryretr=True, trycwd=False),
    ),
    (
        "drwxrwxr-x               folder        2 May 10  1996 network",
        dict(name="network", tryretr=False, trycwd=True),
    ),
    # MultiNet (some spaces removed from examples)
    (
        "00README.TXT;1      2 30-DEC-1996 17:44 [SYSTEM] (RWED,RWED,RE,RE)",
        dict(name="00README.TXT", tryretr=True, trycwd=False),
    ),
    (
        "CORE.DIR;1          1  8-SEP-1996 16:09 [SYSTEM] (RWE,RWE,RE,RE)",
        dict(name="CORE", tryretr=False, trycwd=True),
    ),
    # and non-MutliNet VMS:
    (
        "CII-MANUAL.TEX;1  213/216  29-JAN-1996 03:33:12  [ANONYMOU,ANONYMOUS]   (RWED,RWED,,)",
        dict(name="CII-MANUAL.TEX", tryretr=True, trycwd=False),
    ),
    # MS-DOS format
    (
        "04-27-00  09:09PM       <DIR>          licensed",
        dict(name="licensed", tryretr=False, trycwd=True),
    ),
    (
        "07-18-00  10:16AM       <DIR>          pub",
        dict(name="pub", tryretr=False, trycwd=True),
    ),
    (
        "04-14-00  03:47PM                  589 readme.htm",
        dict(name="readme.htm", tryretr=True, trycwd=False),
    ),
    # Some useless lines, safely ignored:
    ("Total of 11 Files, 10966 Blocks.", None),  # (VMS)
    ("total 14786", None),  # (UNIX)
    ("DISK$ANONFTP:[ANONYMOUS]", None),  # (VMS)
    ("Directory DISK$PCSA:[ANONYM]", None),  # (VMS)
    ("", None),
)


class TestFtpparse(unittest.TestCase):
    """
    Test FTP LIST line parsing.
    """

    def test_ftpparse(self):
        for line, expected in patterns:
            res = ftpparse(line)
            self.assertEqual(
                expected, res, f"got {res!r}\nexpected {expected!r}\n{line!r}"
            )



# Copyright (C) 2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test http stuff with httpbin.org.
"""
import re
from tests import need_network, running_in_ci
from . import LinkCheckTest


def get_httpbin_url(path):
    """Get httpbin URL."""
    if running_in_ci():
        host = "localhost:8080"
    else:
        host = "httpbin.org"
    return f"http://{host}{path}"


class TestHttpbin(LinkCheckTest):
    """Test http:// link redirection checking."""

    @need_network
    def test_http_link(self):
        linkurl = "https://linkchecker.github.io"
        nlinkurl = self.norm(linkurl)
        url = get_httpbin_url("/response-headers?Link=<%s>;rel=previous" % linkurl)
        nurl = self.norm(url)
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            "warning The URL with content type 'application/json' is not parseable.",
            "valid",
            "url %s" % linkurl,
            "cache key %s" % nlinkurl,
            "real url %s" % nlinkurl,
            "name Link: header previous",
            "valid",
        ]
        self.direct(url, resultlines, recursionlevel=1)

    @need_network
    def test_basic_auth(self):
        user = "testuser"
        password = "testpassword"
        url = get_httpbin_url(f"/basic-auth/{user}/{password}")
        nurl = self.norm(url)
        entry = dict(user=user, password=password, pattern=re.compile(r".*"))
        confargs = dict(authentication=[entry])
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            "valid",
        ]
        self.direct(url, resultlines, confargs=confargs)

    @need_network
    def test_http_refresh_header(self):
        linkurl = "https://linkchecker.github.io"
        nlinkurl = self.norm(linkurl)
        url = get_httpbin_url("/response-headers?Refresh=5;url=%s" % linkurl)
        nurl = self.norm(url)
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            "warning The URL with content type 'application/json' is not parseable.",
            "valid",
            "url %s" % linkurl,
            "cache key %s" % nlinkurl,
            "real url %s" % nlinkurl,
            "name Refresh: header",
            "valid",
        ]
        self.direct(url, resultlines, recursionlevel=1)

    @need_network
    def test_http_content_location_header(self):
        linkurl = "https://linkchecker.github.io"
        nlinkurl = self.norm(linkurl)
        url = get_httpbin_url("/response-headers?Content-Location=%s" % linkurl)
        nurl = self.norm(url)
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            "warning The URL with content type 'application/json' is not parseable.",
            "valid",
            "url %s" % linkurl,
            "cache key %s" % nlinkurl,
            "real url %s" % nlinkurl,
            "name Content-Location: header",
            "valid",
        ]
        self.direct(url, resultlines, recursionlevel=1)



# Copyright (C) 2001-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Find link tags in HTML text.
"""

import re

from .srcsetparse import parse_srcset
from .. import strformat, log, LOG_CHECK, url as urlutil

unquote = strformat.unquote

# HTML4/5 link tags
# ripped mainly from HTML::Tagset.pm with HTML5 added
LinkTags = {
    'a': ['href'],
    'applet': ['archive', 'src'],
    'area': ['href'],
    'audio': ['src'],  # HTML5
    'bgsound': ['src'],
    'blockquote': ['cite'],
    'body': ['background'],
    'button': ['formaction'],  # HTML5
    'del': ['cite'],
    'embed': ['pluginspage', 'src'],
    'form': ['action'],
    'frame': ['src', 'longdesc'],
    'head': ['profile'],
    'html': ['manifest'],  # HTML5
    'iframe': ['src', 'longdesc'],
    'ilayer': ['background'],
    'img': ['src', 'lowsrc', 'longdesc', 'usemap', 'srcset'],
    'input': ['src', 'usemap', 'formaction'],
    'ins': ['cite'],
    'isindex': ['action'],
    'layer': ['background', 'src'],
    'link': ['href'],
    'meta': ['content', 'href'],
    'object': ['classid', 'data', 'archive', 'usemap', 'codebase'],
    'q': ['cite'],
    'script': ['src'],
    'source': ['src'],  # HTML5
    'table': ['background'],
    'td': ['background'],
    'th': ['background'],
    'tr': ['background'],
    'track': ['src'],  # HTML5
    'video': ['src'],  # HTML5
    'xmp': ['href'],
    None: ['style', 'itemtype'],
}

# HTML anchor tags
AnchorTags = {
    'a': ['name'],
    None: ['id'],
}

# WML tags
WmlTags = {
    'a': ['href'],
    'go': ['href'],
    'img': ['src'],
}


# matcher for <meta http-equiv=refresh> tags
refresh_re = re.compile(r"(?i)^\d+;\s*url=(?P<url>.+)$")

_quoted_pat = r"('[^']+'|\"[^\"]+\"|[^\)\s]+)"
css_url_re = re.compile(r"url\(\s*(?P<url>%s)\s*\)" % _quoted_pat)

# Note that swf_url_re, unlike all other regular expressions here, is meant
# to match byte strings.  Yes, we're scraping binary SWF data for anything
# that looks like a URL.  What did you expect, a full SWF format decoder?
swf_url_re = re.compile(b"(?i)%s" % urlutil.safe_url_pattern.encode('ascii'))

c_comment_re = re.compile(r"/\*.*?\*/", re.DOTALL)


def strip_c_comments(text):
    """Remove C/CSS-style comments from text. Note that this method also
    deliberately removes comments inside of strings."""
    return c_comment_re.sub('', text)


def is_meta_url(attr, attrs):
    """Check if the meta attributes contain a URL."""
    res = False
    if attr == "content":
        equiv = attrs.get('http-equiv', '').lower()
        scheme = attrs.get('scheme', '').lower()
        res = equiv in ('refresh',) or scheme in ('dcterms.uri',)
    if attr == "href":
        rel = attrs.get('rel', '').lower()
        res = rel in ('shortcut icon', 'icon')
    return res


def is_form_get(attr, attrs):
    """Check if this is a GET form action URL."""
    res = False
    if attr == "action":
        method = attrs.get('method', '').lower()
        res = method != 'post'
    return res


class LinkFinder:
    """Find HTML links, and apply them to the callback function with the
    format (url, lineno, column, name, codebase)."""

    def __init__(self, callback, tags):
        """Store content in buffer and initialize URL list."""
        self.callback = callback
        # set universal tag attributes using tagname None
        self.universal_attrs = set(tags.get(None, []))
        self.tags = dict()
        for tag, attrs in tags.items():
            self.tags[tag] = set(attrs)
            # add universal tag attributes
            self.tags[tag].update(self.universal_attrs)
        self.base_ref = ''

    def html_element(self, tag, attrs, element_text, lineno, column):
        """Search for links and store found URLs in a list."""
        log.debug(LOG_CHECK, "LinkFinder tag %s attrs %s", tag, attrs)
        log.debug(LOG_CHECK, "line %d col %d", lineno, column)
        if tag == "base" and not self.base_ref:
            self.base_ref = attrs.get("href", '')
        tagattrs = self.tags.get(tag, self.universal_attrs)
        # parse URLs in tag (possibly multiple URLs in CSS styles)
        for attr in sorted(tagattrs.intersection(attrs)):
            if tag == "meta" and not is_meta_url(attr, attrs):
                continue
            if tag == "form" and not is_form_get(attr, attrs):
                continue
            # name of this link
            name = self.get_link_name(tag, attrs, attr, element_text)
            # possible codebase
            base = ''
            if tag == 'applet':
                base = attrs.get('codebase', '')
            if not base:
                base = self.base_ref
            # note: value can be None
            value = attrs.get(attr)
            if tag == 'link' and (rel := attrs.get('rel', '').lower()) \
                    and ('dns-prefetch' in rel or 'preconnect' in rel):
                if ':' in value:
                    value = value.split(':', 1)[1]
                value = 'dns:' + value.rstrip('/')
            # parse tag for URLs
            self.parse_tag(tag, attr, value, name, base, lineno, column)
        log.debug(LOG_CHECK, "LinkFinder finished tag %s", tag)

    def get_link_name(self, tag, attrs, attr, name=None):
        """Parse attrs for link name. Return name of link."""
        if tag == 'a' and attr == 'href':
            if not name:
                name = attrs.get('title', '')
        elif tag == 'img':
            name = attrs.get('alt', '')
            if not name:
                name = attrs.get('title', '')
        else:
            name = ""
        return name

    def parse_tag(self, tag, attr, value, name, base, lineno, column):
        """Add given url data to url list."""
        assert isinstance(tag, str), repr(tag)
        assert isinstance(attr, str), repr(attr)
        assert isinstance(name, str), repr(name)
        assert isinstance(base, str), repr(base)
        assert isinstance(value, str) or value is None, repr(value)
        # look for meta refresh
        if tag == 'meta' and value:
            mo = refresh_re.match(value)
            if mo:
                self.found_url(mo.group("url"), name, base, lineno, column)
            elif attr != 'content':
                self.found_url(value, name, base, lineno, column)
        elif attr == 'style' and value:
            for mo in css_url_re.finditer(value):
                url = unquote(mo.group("url"), matching=True)
                self.found_url(url, name, base, lineno, column)
        elif attr == 'archive':
            for url in value.split(','):
                self.found_url(url, name, base, lineno, column)
        elif attr == 'srcset':
            for url in parse_srcset(value):
                self.found_url(url, name, base, lineno, column)
        else:
            self.found_url(value, name, base, lineno, column)

    def found_url(self, url, name, base, lineno, column):
        """Add newly found URL to queue."""
        assert isinstance(url, str) or url is None, repr(url)
        self.callback(url, line=lineno, column=column, name=name, base=base)


def find_links(soup, callback, tags):
    """Parse into content and search for URLs to check.
    When a URL is found it is passed to the supplied callback.
    """
    lf = LinkFinder(callback, tags)
    for element in soup.find_all(True):
        lf.html_element(
            element.name,
            element.attrs,
            element.text.strip(),
            element.sourceline,
            None if element.sourcepos is None else element.sourcepos + 1,
        )



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test url build method from url data objects.
"""

import unittest
import linkcheck.configuration
import linkcheck.director
import linkcheck.checker.urlbase
from linkcheck.checker import get_url_from


def get_test_aggregate():
    """
    Initialize a test configuration object.
    """
    config = linkcheck.configuration.Configuration()
    config["logger"] = config.logger_new("none")
    return linkcheck.director.get_aggregate(config)


class TestUrlBuild(unittest.TestCase):
    """
    Test url building.
    """

    def test_http_build(self):
        parent_url = "http://localhost:8001/tests/checker/data/http.html"
        base_url = "http://foo"
        recursion_level = 0
        aggregate = get_test_aggregate()
        o = get_url_from(base_url, recursion_level, aggregate, parent_url=parent_url)
        o.build_url()
        self.assertEqual(o.url, "http://foo")

    def test_urljoin(self):
        parent_url = "http://localhost:8001/test"
        base_url = ";param=value"
        res = linkcheck.checker.urlbase.urljoin(parent_url, base_url)
        self.assertEqual(res, "http://localhost:8001/;param=value")

    def test_urljoin_file(self):
        parent_url = "file:///a/b.html"
        base_url = "?c=d"
        recursion_level = 0
        aggregate = get_test_aggregate()
        o = get_url_from(base_url, recursion_level, aggregate, parent_url=parent_url)
        o.build_url()
        self.assertEqual(o.url, parent_url)

    def test_http_build2(self):
        parent_url = "http://example.org/test?a=b&c=d"
        base_url = "#usemap"
        recursion_level = 0
        aggregate = get_test_aggregate()
        o = get_url_from(base_url, recursion_level, aggregate, parent_url=parent_url)
        o.build_url()
        self.assertEqual(o.url, parent_url + base_url)



# Copyright (C) 2020-2023 Chris Mayo
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test director.aggregator.Aggregate.visit_loginurl().
This includes the retrieval and search of a login form and posting credentials.
"""
import re
import urllib.parse

from .httpserver import HttpServerTest, StoppableHttpRequestHandler
from . import get_test_aggregate


class LoginPostHandler(StoppableHttpRequestHandler):
    def do_POST(self):
        length = int(self.headers.get("content-length"))
        form_data = self.rfile.read(length).decode()
        self.send_response(200)
        for name, value in urllib.parse.parse_qsl(form_data):
            self.send_header("Set-Cookie", f"{name}={value}")
        self.end_headers()


class TestLoginUrl(HttpServerTest):
    """Test loginurl retrieval, search and posting credentials."""

    def __init__(self, methodName="runTest"):
        super().__init__(methodName=methodName)
        self.handler = LoginPostHandler

    def visit_loginurl(self, page, user=None, password=None, extrafields=False):
        confargs = {}
        confargs["loginurl"] = self.get_url(page)
        if extrafields:
            confargs["loginextrafields"] = {"extra_field": "default"}
        confargs["authentication"] = [
            {
                "user": user,
                "password": password,
                "pattern": re.compile("^http://localhost.*"),
            },
        ]

        aggregate = get_test_aggregate(confargs, {"expected": ""})
        aggregate.visit_loginurl()

        return aggregate.cookies

    def test_loginurl(self):
        cookies = self.visit_loginurl(
            "loginform.html", "test_user", "test_password", True
        )

        self.assertEqual(cookies["login"], "test_user")
        self.assertEqual(cookies["password"], "test_password")
        self.assertEqual(cookies["extra_field"], "default")

    def test_loginurl_user(self):
        cookies = self.visit_loginurl("loginform_user.html", "test_user")

        self.assertEqual(cookies["login"], "test_user")

    def test_login_password(self):
        cookies = self.visit_loginurl(
            "loginform_password.html", password="test_password"
        )

        self.assertEqual(cookies["password"], "test_password")



# Copyright (C) 2006-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Logger for aggregator instances"""
import threading
import _thread

from ..decorators import synchronized

_lock = threading.Lock()


class Logger:
    """Thread safe multi-logger class used by aggregator instances."""

    def __init__(self, config):
        """Initialize basic logging variables."""
        self.loggers = [config['logger']]
        self.loggers.extend(config['fileoutput'])
        self.verbose = config["verbose"]
        self.warnings = config["warnings"]

    def start_log_output(self):
        """
        Start output of all configured loggers.
        """
        for logger in self.loggers:
            logger.start_output()

    def end_log_output(self, **kwargs):
        """
        End output of all configured loggers.
        """
        for logger in self.loggers:
            logger.end_output(**kwargs)

    def do_print(self, url_data):
        """Determine if URL entry should be logged or not."""
        if self.verbose:
            return True
        if self.warnings and url_data.warnings:
            return True
        return not url_data.valid

    @synchronized(_lock)
    def log_url(self, url_data):
        """Send new url to all configured loggers."""
        self.check_active_loggers()
        do_print = self.do_print(url_data)
        # Only send a transport object to the loggers, not the complete
        # object instance.
        for log in self.loggers:
            log.log_filter_url(url_data, do_print)

    @synchronized(_lock)
    def log_internal_error(self):
        """Document that an internal error occurred."""
        for logger in self.loggers:
            logger.log_internal_error()

    def check_active_loggers(self):
        """Check if all loggers are deactivated due to I/O errors."""
        for logger in self.loggers:
            if logger.is_active:
                break
        else:
            _thread.interrupt_main()



# Copyright (C) 2005-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Simple decorators (usable in Python >= 2.4).

Example::

    @synchronized(thread.allocate_lock())
    def f():
        "Synchronized function"
        print("i am synchronized:", f, f.__doc__)

    @deprecated
    def g():
        "this function is deprecated"
        pass

    @notimplemented
    def h():
        "todo"
        pass

"""
import warnings
import signal
import os
import sys
import time


def update_func_meta(fake_func, real_func):
    """Set meta information (eg. __doc__) of fake function to that
    of the real function.
    @return fake_func
    """
    fake_func.__module__ = real_func.__module__
    fake_func.__name__ = real_func.__name__
    fake_func.__doc__ = real_func.__doc__
    fake_func.__dict__.update(real_func.__dict__)
    return fake_func


def deprecated(func):
    """A decorator which can be used to mark functions as deprecated.
    It emits a warning when the function is called."""

    def newfunc(*args, **kwargs):
        """Print deprecated warning and execute original function."""
        warnings.warn(
            "Call to deprecated function %s." % func.__name__,
            category=DeprecationWarning,
        )
        return func(*args, **kwargs)

    return update_func_meta(newfunc, func)


def signal_handler(signal_number):
    """From http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/410666

    A decorator to set the specified function as handler for a signal.
    This function is the 'outer' decorator, called with only the
    (non-function) arguments.
    If signal_number is not a valid signal (for example signal.SIGN),
    no handler is set.
    """
    # create the 'real' decorator which takes only a function as an argument
    def newfunc(function):
        """Register function as signal handler."""
        # note: actually the kill(2) function uses the signal number of 0
        # for a special case, but for signal(2) only positive integers
        # are allowed
        is_valid_signal = 0 < signal_number < signal.NSIG
        if is_valid_signal and os.name == 'posix':
            signal.signal(signal_number, function)
        return function

    return newfunc


def synchronize(lock, func, log_duration_secs=0):
    """Return synchronized function acquiring the given lock."""

    def newfunc(*args, **kwargs):
        """Execute function synchronized."""
        t = time.time()
        with lock:
            duration = time.time() - t
            if duration > log_duration_secs > 0:
                print(
                    "WARN:",
                    func.__name__,
                    "locking took %0.2f seconds" % duration,
                    file=sys.stderr,
                )
            return func(*args, **kwargs)

    return update_func_meta(newfunc, func)


def synchronized(lock):
    """A decorator calling a function with acquired lock."""
    return lambda func: synchronize(lock, func)


def notimplemented(func):
    """Raises a NotImplementedError if the function is called."""

    def newfunc(*args, **kwargs):
        """Raise NotImplementedError"""
        co = func.func_code
        attrs = (co.co_name, co.co_filename, co.co_firstlineno)
        raise NotImplementedError("function %s at %s:%d is not implemented" % attrs)

    return update_func_meta(newfunc, func)


def timeit(func, log, limit):
    """Print execution time of the function. For quick'n'dirty profiling."""

    def newfunc(*args, **kwargs):
        """Execute function and print execution time."""
        t = time.time()
        res = func(*args, **kwargs)
        duration = time.time() - t
        if duration > limit:
            print(func.__name__, "took %0.2f seconds" % duration, file=log)
            print(args, file=log)
            print(kwargs, file=log)
        return res

    return update_func_meta(newfunc, func)


def timed(log=sys.stderr, limit=2.0):
    """Decorator to run a function with timing info."""
    return lambda func: timeit(func, log, limit)


class curried:
    """Decorator that returns a function that keeps returning functions
    until all arguments are supplied; then the original function is
    evaluated."""

    def __init__(self, func, *a):
        """Store function and arguments."""
        self.func = func
        self.args = a

    def __call__(self, *a):
        """If all arguments function arguments are supplied, call it.
        Else return another curried object."""
        args = self.args + a
        if len(args) < self.func.func_code.co_argcount:
            return curried(self.func, *args)
        else:
            return self.func(*args)



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test mail checking of bad mail addresses.
"""
from . import MailTest


class TestMailBad(MailTest):
    """Test mailto: link checking."""

    def test_error_mail(self):
        # too long or too short
        self.mail_error("mailto:@")
        self.mail_error("mailto:@example.org")
        self.mail_error("mailto:a@")
        self.mail_error("mailto:%s@example.org" % ("a" * 65))
        self.mail_error("mailto:a@%s.com" % ("a" * 64))
        # local part quoted
        self.mail_error('mailto:"a""@example.com', cache_key="mailto:a")
        self.mail_error('mailto:""a"@example.com', cache_key='mailto:""a"@example.com')
        self.mail_error('mailto:"a\\"@example.com', cache_key='mailto:a"@example.com')
        # local part unquoted
        self.mail_error("mailto:.a@example.com")
        self.mail_error("mailto:a.@example.com")
        self.mail_error("mailto:a..b@example.com")
        # domain part
        self.mail_error("mailto:a@a_b.com")
        self.mail_error("mailto:a@example.com.")
        self.mail_error("mailto:a@example.com.111")
        self.mail_error("mailto:a@example..com")
        # other
        # ? extension forbidden in <> construct
        self.mail_error(
            "mailto:Bastian Kleineidam <calvin@users.sourceforge.net?foo=bar>",
            cache_key="mailto:calvin@users.sourceforge.net?foo=bar",
        )



# Copyright (C) 2008-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test network functions.
"""

import unittest
from tests import need_network, need_linux
from linkcheck.network import iputil


class TestNetwork(unittest.TestCase):
    """Test network functions."""

    @need_network
    @need_linux
    def test_iputils(self):
        # note: need a hostname whose reverse lookup of the IP is the same host
        host = "dinsdale.python.org"
        ips = iputil.resolve_host(host)
        self.assertTrue(len(ips) > 0)



# Copyright (C) 2012-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test URL length checks.
"""
from . import LinkCheckTest
from linkcheck.checker.const import URL_MAX_LENGTH


class TestURLLength(LinkCheckTest):
    """
    Test URL lengths.
    """

    def test_url_warn(self):
        url = "http://www.example.org/" + ("a" * URL_MAX_LENGTH)
        attrs = self.get_attrs(url=url)
        attrs["nurl"] = self.norm(url)
        resultlines = [
            "url %(nurl)s" % attrs,
            "cache key %(nurl)s" % attrs,
            "real url %(nurl)s" % attrs,
            "warning URL length %d is longer than %d." % (len(url), URL_MAX_LENGTH),
            "error",
        ]
        self.direct(url, resultlines)



# Copyright (C) 2020 Chris Mayo
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test VirusCheck plugin URL based check.
"""
from . import LinkCheckTest
from .. import need_clamav


class TestVirusCheck(LinkCheckTest):
    """
    Test virus scan of page content.
    """

    @need_clamav
    def test_viruscheck(self):
        confargs = {"enabledplugins": ["VirusCheck"]}
        check_file = "base1.html"
        attrs = self.get_attrs()
        attrs.update({"check_file": check_file})
        url = "file://%(curdir)s/%(datadir)s/%(check_file)s" % attrs
        rurl = self.get_url(check_file)
        resultlines = [
            "url %s" % url,
            "cache key %s" % url,
            "real url %s" % url,
            "name %s" % rurl,
            "info No viruses in data found.",
            "valid",
        ]
        self.direct(rurl, resultlines, confargs=confargs)



import os
import sys
sys.path.insert(0, os.path.abspath('../..'))

# -- Project information -----------------------------------------------------

import linkcheck.configuration

project = 'LinkChecker'
copyright = linkcheck.configuration.Copyright.split("Copyright (C) ")[1]
version = linkcheck.configuration.Version
release = version

# -- General configuration ---------------------------------------------------

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.autosummary',
    'sphinx.ext.extlinks',
    'sphinx.ext.graphviz',
    'sphinx.ext.viewcode',
    'sphinx_epytext',
    'sphinx_rtd_theme',
    'sphinx_sitemap',
]

locale_dirs = ['../i18n/locales']

templates_path = ['_templates']

today_fmt = '%B %d, %Y'

# -- Options for HTML output -------------------------------------------------

html_favicon = 'images/favicon.ico'

html_logo = 'images/logo128x128.png'

html_theme = 'sphinx_rtd_theme'

html_theme_options = {
    'collapse_navigation': False
}

# only use :manpage: within man pages
manpages_url = '{page}.html'

# -- Options for man output -------------------------------------------------

man_pages = [
    (
     'man/linkchecker', 'linkchecker',
     'Kommandozeilenprogramm zum Prüfen von HTML Dokumenten und '
     'Webseiten auf ungültige Verknüpfungen'
     if tags.has('de') else
     'command line client to check HTML documents and websites for broken links',
     ['Bastian Kleineidam <bastian.kleineidam@web.de>'], 1),
    (
     'man/linkcheckerrc', 'linkcheckerrc',
     'Konfigurationsdatei für LinkChecker'
     if tags.has('de') else
     'configuration file for LinkChecker',
     ['Bastian Kleineidam <bastian.kleineidam@web.de>'], 5),
]

# -- Extension configuration -------------------------------------------------

autoclass_content = 'both'

autodoc_default_options = {
    'members': True,
    'undoc-members': True,
    'show-inheritance': True,
}

autodoc_member_order = 'groupwise'

autosectionlabel_prefix_document = True

autosummary_mock_imports = ['ctypes', 'linkcheck.__main__']

extlinks = {'pypi': ('https://pypi.org/project/%s/', '%s')}

graphviz_output_format = 'svg'

# sitemap
html_baseurl = 'https://linkchecker.github.io/linkchecker/'

sitemap_locales = [None]

sitemap_url_scheme = "{link}"

# -- Mock --------------------------------------------------------------------

import linkcheck.logger
linkcheck.logger.failures.FailuresLogger.LoggerArgs = {
    'filename': '$XDG_DATA_HOME/linkchecker/failures'}



# Copyright (C) 2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Module for plugin management.
"""
from .. import loader, log, LOG_PLUGIN
from ..decorators import notimplemented


class _PluginBase:
    """Basic plugin class featuring plugin identification and
    helper functions."""

    def __init__(self, config):
        """Add plugin-specific configuration."""
        pass

    @notimplemented
    def applies_to(self, url_data, **kwargs):
        """See if url_data should be handled by this plugin."""
        pass

    @notimplemented
    def check(self, url_data):
        """Common check method run for all plugins."""
        pass

    @classmethod
    def read_config(cls, configparser):
        """Read configuration file options."""
        pass


class _ConnectionPlugin(_PluginBase):
    """Plugins run after connection checks."""

    pass


class _ContentPlugin(_PluginBase):
    """Plugins run for valid URLs with content."""

    pass


class _ParserPlugin(_PluginBase):
    """Plugins run for valid URLs to parse their contents."""

    pass


def get_plugin_modules(folders):
    """Get plugin modules for given folders."""
    for folder in folders:
        yield from loader.get_folder_modules(folder, 'linkcheck.dummy')
    yield from loader.get_package_modules('plugins', __path__)


def get_plugin_classes(modules):
    """Get plugin classes for given modules."""
    classes = (_ConnectionPlugin, _ContentPlugin, _ParserPlugin)
    return loader.get_plugins(modules, classes)


class PluginManager:
    """Manage all connection and content plugins."""

    def __init__(self, config):
        """Load enabled plugins."""
        self.connection_plugins = []
        self.content_plugins = []
        self.parser_plugins = []
        folders = config["pluginfolders"]
        modules = get_plugin_modules(folders)
        self.load_modules(modules, config)

    def load_modules(self, modules, config):
        """Load plugin modules."""
        for pluginclass in get_plugin_classes(modules):
            name = pluginclass.__name__
            if name in config["enabledplugins"]:
                if issubclass(pluginclass, _ConnectionPlugin):
                    log.debug(LOG_PLUGIN, "Enable connection plugin %s", name)
                    self.connection_plugins.append(pluginclass(config[name]))
                elif issubclass(pluginclass, _ContentPlugin):
                    log.debug(LOG_PLUGIN, "Enable content plugin %s", name)
                    self.content_plugins.append(pluginclass(config[name]))
                elif issubclass(pluginclass, _ParserPlugin):
                    log.debug(LOG_PLUGIN, "Enable parser plugin %s", name)
                    self.parser_plugins.append(pluginclass(config[name]))
                else:
                    raise ValueError("Invalid plugin class %s" % pluginclass)

    def run_connection_plugins(self, url_data):
        """Run all connection plugins."""
        run_plugins(self.connection_plugins, url_data)

    def run_content_plugins(self, url_data):
        """Run all content plugins."""
        run_plugins(self.content_plugins, url_data)

    def run_parser_plugins(self, url_data, pagetype):
        """Run parser plugins for given pagetype."""
        run_plugins(
            self.parser_plugins, url_data, stop_after_match=True, pagetype=pagetype
        )


def run_plugins(plugins, url_data, stop_after_match=False, **kwargs):
    """Run the check(url_data) method of given plugins."""
    for plugin in plugins:
        log.debug(LOG_PLUGIN, "Run plugin %s", plugin.__class__.__name__)
        if plugin.applies_to(url_data, **kwargs):
            plugin.check(url_data)
            if stop_after_match:
                break



# Copyright (C) 2020 Chris Mayo
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test UrlBase.build_url()
"""
import linkcheck.configuration
import linkcheck.director

from . import get_url_from
from .. import TestBase


class TestBuildUrl(TestBase):
    """Test parsing of URLs by UrlBase.build_url()."""

    def test_build_url(self):
        config = linkcheck.configuration.Configuration()
        aggregate = linkcheck.director.get_aggregate(config)
        url = "https://user:password@host:1234/path/?key1=value1&key2=value2#fragment"
        url_data = get_url_from(url, 0, aggregate)
        self.assertEqual(url_data.scheme, "https")
        self.assertEqual(url_data.userinfo, "user:password")
        self.assertEqual(url_data.host, "host")
        self.assertEqual(url_data.port, 1234)
        self.assertEqual(url_data.anchor, "fragment")

        self.assertEqual(url_data.get_user_password(), ("user", "password"))

        # user without password
        url = "https://user@host/"
        url_data = get_url_from(url, 0, aggregate)
        self.assertEqual(url_data.userinfo, "user")

        # password without user
        url = "https://:password@host/"
        url_data = get_url_from(url, 0, aggregate)
        self.assertEqual(url_data.userinfo, ":password")

        # invalid port
        url = "https://host:abc/"
        url_data = get_url_from(url, 0, aggregate)
        self.assertFalse(url_data.valid)
